@@ -60,6 +60,8 @@ namespace FCam { namespace N900 {
         // make the semaphore
         sem_init(&actionQueueSemaphore, 0, 0);
 
+        lastGoodShot.wanted = false;
+
         pipelineFlush = true;
     }
 
@@ -216,7 +218,15 @@ namespace FCam { namespace N900 {
             }
             tickSetter(Time(t));
         }
-
+	// When shutting down, make sure we leave V4L2 with the last requested Shot's 
+	// settings, so that they can be used by later programs. Otherwise, the bubbles
+	// produced during shutdown will clobber them. Gain and WB are not written by bubbles
+	// so don't need to be dealt with here.
+        if (lastGoodShot.wanted == true) {
+            v4l2Sensor->setFrameTime(lastGoodShot.frameTime);
+            v4l2Sensor->setExposure(lastGoodShot.exposure);
+            v4l2Sensor->setGain(lastGoodShot.gain);
+        }
     }
 
     void Daemon::tickSetter(Time hs_vs) {
@@ -242,34 +252,36 @@ namespace FCam { namespace N900 {
             dprintf(4, "Setter: setting gain and WB\n");
             // set the gain and predicted done time on the pending request
             // and then push it onto the handler's input queue and the v4l2 buffer queue
-
-            if (req->shot().gain != current._shot.gain) {
-                v4l2Sensor->setGain(req->shot().gain);
-                current._shot.gain = req->shot().gain;
-            }
-            current.gain = req->gain = v4l2Sensor->getGain();
-
-            if (req->shot().colorMatrix().size() != 12) {
-                if (req->shot().whiteBalance != current._shot.whiteBalance) {                
-                    int wb = req->shot().whiteBalance;
-                    
-                    // Very lenient sanity checks - restricting the range is up to the auto-wb algorithm
-                    if (wb < 0) wb = 0; // a black-body radiator at absolute zero.
-                    if (wb > 25000) wb = 25000; // A type 'O' star.
-                    
-                    float matrix[12];
-                    sensor->platform().rawToRGBColorMatrix(wb, matrix);
-                    v4l2Sensor->setWhiteBalance(matrix);
-                    
-                    current._shot.whiteBalance = req->shot().whiteBalance;
-                    current.whiteBalance = wb;            
-                }
-                req->whiteBalance = current.whiteBalance;
-            } else {
-                // custom WB - always set it                
-                v4l2Sensor->setWhiteBalance(&req->shot().colorMatrix()[0]);
-                current._shot.setColorMatrix(&req->shot().colorMatrix()[0]);
-            }
+	    // Only update gain/wb for wanted requests, bubbles shouldn't change these.
+	    if (req->shot().wanted == true) {
+		if (req->shot().gain != current._shot.gain) {
+		    v4l2Sensor->setGain(req->shot().gain);
+		    current._shot.gain = req->shot().gain;
+		}
+		current.gain = req->gain = v4l2Sensor->getGain();
+		
+		if (req->shot().colorMatrix().size() != 12) {
+		    if (req->shot().whiteBalance != current._shot.whiteBalance) {                
+			int wb = req->shot().whiteBalance;
+			
+			// Very lenient sanity checks - restricting the range is up to the auto-wb algorithm
+			if (wb < 0) wb = 0; // a black-body radiator at absolute zero.
+			if (wb > 25000) wb = 25000; // A type 'O' star.
+			
+			float matrix[12];
+			sensor->platform().rawToRGBColorMatrix(wb, matrix);
+			v4l2Sensor->setWhiteBalance(matrix);
+			
+			current._shot.whiteBalance = req->shot().whiteBalance;
+			current.whiteBalance = wb;            
+		    }
+		    req->whiteBalance = current.whiteBalance;
+		} else {
+		    // custom WB - always set it                
+		    v4l2Sensor->setWhiteBalance(&req->shot().colorMatrix()[0]);
+		    current._shot.setColorMatrix(&req->shot().colorMatrix()[0]);
+		}
+	    }
 
             // Predict when this frame will be done. It should be HS_VS +
             // the readout time for the previous frame + the frame time
@@ -340,6 +352,15 @@ namespace FCam { namespace N900 {
             dprintf(4, "Setter: grabbing next request\n");
             // There's a real request for us to handle
             req = requestQueue.front();
+            // Keep track of the last good shot parameters so we can set those on exit
+            // since bubbles will clobber the V4L2 state on shutdown otherwise.
+            // This allows us to be nice to other programs hoping to use our last settings.
+            if (req->_shot.wanted == true) {
+                lastGoodShot.wanted = true;
+                lastGoodShot.exposure = req->_shot.exposure;
+                lastGoodShot.gain = req->_shot.gain;
+                lastGoodShot.frameTime = req->_shot.frameTime;
+            }
         } else {
             dprintf(4, "Setter: inserting a bubble\n");
             // There are no pending requests, push a bubble into the
