@@ -112,6 +112,9 @@ namespace FCam { namespace N900 {
         struct v4l2_streamparm parm;
         memset(&parm, 0, sizeof(struct v4l2_streamparm));
         parm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
+        parm.parm.capture.timeperframe.numerator = 0;
+        parm.parm.capture.timeperframe.denominator = 1000000;
+        /*
         if (m.height <= 960) {
             parm.parm.capture.timeperframe.numerator = 33000;
             parm.parm.capture.timeperframe.denominator = 1000000;
@@ -119,6 +122,7 @@ namespace FCam { namespace N900 {
             parm.parm.capture.timeperframe.numerator = 77000;
             parm.parm.capture.timeperframe.denominator = 1000000;
         }
+        */
         if (ioctl(fd, VIDIOC_S_PARM, &parm) < 0) {
             error(Event::DriverError, "VIDIOC_S_PARM");
             return;
@@ -566,6 +570,8 @@ namespace FCam { namespace N900 {
     void V4L2Sensor::setExposure(int e) {
         if (state == CLOSED) return;
 
+        dprintf(3, "Setting exposure to %d\n", e);
+
         struct v4l2_control ctrl;
         ctrl.id = V4L2_CID_EXPOSURE;
         ctrl.value = e;
@@ -586,6 +592,8 @@ namespace FCam { namespace N900 {
             error(Event::DriverError, "VIDIOC_G_CTRL: %s", strerror(errno));
             return -1;
         }       
+
+        dprintf(3, "Getting exposure: %d\n", ctrl.value);
         
         return ctrl.value;
     }
@@ -598,6 +606,7 @@ namespace FCam { namespace N900 {
         struct v4l2_control ctrl;
         ctrl.id = V4L2_CID_FRAME_TIME;
         ctrl.value = e;
+        dprintf(3, "Setting frame time to %d\n", e);
         if (ioctl(fd, VIDIOC_S_CTRL, &ctrl) < 0) {
             error(Event::DriverError, "VIDIOC_S_CTRL: %s", strerror(errno));
             return;
@@ -614,6 +623,8 @@ namespace FCam { namespace N900 {
             error(Event::DriverError, "VIDIOC_G_CTRL: %s", strerror(errno));
             return -1;
         }       
+
+        dprintf(3, "Getting frame time: %d\n", ctrl.value);
         
         return ctrl.value;
     }
