@@ -4,7 +4,7 @@
 #include <string.h>
 #include <time.h>
 #include <sched.h>
-#include <math.h>
+#include <cmath>
 #include <errno.h>
 
 #include <FCam/processing/DNG.h>
@@ -12,20 +12,30 @@
 #include <FCam/processing/Demosaic.h>
 #include <FCam/Event.h>
 #include <FCam/Frame.h>
+#include <FCam/Platform.h>
 
 #include "TIFF.h"
 #include "../Debug.h"
 
 namespace FCam {
 
+    _DNGFrame::_DNGFrame(): dngFile(NULL) {
+        dngFile = new TiffFile;
+    }
+
+    _DNGFrame::~_DNGFrame() {
+        delete dngFile;
+    }
+
     void _DNGFrame::rawToRGBColorMatrix(int kelvin, float *matrix) const {
         if (dng.numIlluminants == 1) {
             for (int i=0;i < 12; i++) matrix[i] = dng.colorMatrix1[i];
             return;
         }
+        // Linear interpolate using inverse CCT
         float alpha =
-            ((float)kelvin-dng.illuminant1)
-            /(dng.illuminant2 - dng.illuminant1);
+            (1./kelvin-1./dng.illuminant1)
+            /(1./dng.illuminant2 - 1./dng.illuminant1);
         colorMatrixInterpolate(dng.colorMatrix1,
                                dng.colorMatrix2,
                                alpha, matrix);
@@ -33,33 +43,121 @@ namespace FCam {
 
     void _DNGFrame::debug(const char *name) const {
         printf("\tDump of FCam::DNGFrame %s at %llx:\n", name, (long long unsigned)this);
+        printf("\t  Source file name: %s\n", dngFile->filename().c_str());
         printf("\t  Number of calibration illuminants: %d\n", dng.numIlluminants);
         printf("\t  Illuminant 1: %d K, conversion matrix:\n", dng.illuminant1);
         const float *matrix = dng.colorMatrix1;
         printf("\t\t[ [ %5.3f %5.3f %5.3f %5.3f ]\n", matrix[0], matrix[1], matrix[2], matrix[3]);
         printf("\t\t  [ %5.3f %5.3f %5.3f %5.3f ]\n", matrix[4], matrix[5], matrix[6], matrix[7]);
         printf("\t\t  [ %5.3f %5.3f %5.3f %5.3f ]\n", matrix[8], matrix[9], matrix[10], matrix[11]);
-        printf("\t\t  [ %5.3f %5.3f %5.3f %5.3f ] ]\n", matrix[12], matrix[13], matrix[14], matrix[15]);
         if (dng.numIlluminants == 2) {
             printf("\t Illuminant 2: %d K, conversion matrix:\n", dng.illuminant2);
             matrix = dng.colorMatrix2;
             printf("\t\t[ [ %5.3f %5.3f %5.3f %5.3f ]\n", matrix[0], matrix[1], matrix[2], matrix[3]);
             printf("\t\t  [ %5.3f %5.3f %5.3f %5.3f ]\n", matrix[4], matrix[5], matrix[6], matrix[7]);
             printf("\t\t  [ %5.3f %5.3f %5.3f %5.3f ]\n", matrix[8], matrix[9], matrix[10], matrix[11]);
-            printf("\t\t  [ %5.3f %5.3f %5.3f %5.3f ] ]\n", matrix[12], matrix[13], matrix[14], matrix[15]);
         }
-        printf("\t** Dump of DNGFrame thumbnail image data follows\n");
+        printf("\t** Dump of cached DNGFrame thumbnail image data follows\n");
         thumbnail.debug("DNGFrame::thumbnail");
         printf("\t** Dump of base Frame fields follows\n");
         FCam::_Frame::debug("base frame");
     }
 
     DNGFrame::DNGFrame(_DNGFrame *f): FCam::Frame(f) {}
-
+    
+    Image DNGFrame::thumbnail() { 
+        return get()->thumbnail;
+    }
+    
+    const char tiffEPVersion[4] = {1,0,0,0};
     const char understoodDNGVersion[4] = {1,3,0,0};
     const char oldestSupportedDNGVersion[4] = {1,2,0,0};
     const char privateDataPreamble[] = "stanford.fcam.privatedata";
-    const int privateDataVersion = 1;
+    const int privateDataVersion = 2;
+    const int backwardPrivateDataVersion = 2;
+
+    // Notes on privateDataVersion:
+    // version 1:
+    //   Initial release
+    // version 2:
+    //   Added backwards-compatible version field. Readers need to check this to see if they can parse the data.
+    //   All frame data now stored as a key/value set (a serialized TagMap, in other words)
+    //   Frame members have known tag names that are searched for and removed from the TagMap. Anything unknown is left,
+    //   which includes application-added tags and possible frame members from newer implementations.
+
+    void saveDNGPrivateData_v1(const Frame &frame, std::stringstream &privateData,
+                               const std::vector<float> &rawToRGB3000, const std::vector<float> &rawToRGB6500) {
+        privateData << TagValue(frame.exposureStartTime()).toBlob();
+        privateData << TagValue(frame.exposureEndTime()).toBlob();
+        privateData << TagValue(frame.processingDoneTime()).toBlob();
+        privateData << TagValue(frame.exposure()).toBlob();
+        privateData << TagValue(frame.frameTime()).toBlob();
+        privateData << TagValue(frame.gain()).toBlob();
+        privateData << TagValue(frame.whiteBalance()).toBlob();
+        
+        privateData << TagValue(frame.shot().exposure).toBlob();
+        privateData << TagValue(frame.shot().frameTime).toBlob();
+        privateData << TagValue(frame.shot().gain).toBlob();
+        privateData << TagValue(frame.shot().whiteBalance).toBlob();
+        
+        privateData << TagValue(frame.platform().minRawValue()).toBlob();
+        privateData << TagValue(frame.platform().maxRawValue()).toBlob();
+        
+        privateData << TagValue(3000).toBlob();
+        privateData << TagValue(rawToRGB3000).toBlob();
+        privateData << TagValue(6500).toBlob();
+        privateData << TagValue(rawToRGB6500).toBlob();
+        
+        // (not writing histogram/sharpness map)
+        
+        // Write tags
+        for (TagMap::const_iterator it = frame.tags().begin(); it != frame.tags().end(); it++) {
+            privateData << TagValue(it->first) << TagValue(it->second);
+        }
+        
+    }
+
+    void saveDNGPrivateData_v2(const Frame &frame, std::stringstream &privateData, 
+                               const std::vector<float> &rawToRGB3000, const std::vector<float> &rawToRGB6500) {
+
+        // First write backward-compatibility field
+        privateData << TagValue(backwardPrivateDataVersion).toBlob();
+        
+        // Now write everything using a TagMap
+        TagMap frameFields;
+        frameFields["frame.exposureStartTime"] = frame.exposureStartTime();
+        frameFields["frame.exposureEndTime"] = frame.exposureEndTime();
+
+        frameFields["frame.processingDoneTime"] = frame.processingDoneTime();
+        frameFields["frame.exposure"] = frame.exposure();
+        frameFields["frame.frameTime"] = frame.frameTime();
+        frameFields["frame.gain"] = frame.gain();
+        frameFields["frame.whiteBalance"] = frame.whiteBalance();
+        
+        frameFields["frame.shot.exposure"] = frame.shot().exposure;
+        frameFields["frame.shot.frameTime"] = frame.shot().frameTime;
+        frameFields["frame.shot.gain"] = frame.shot().gain;
+        frameFields["frame.shot.whiteBalance"] = frame.shot().whiteBalance;
+        frameFields["frame.shot.colorMatrix"] = frame.shot().colorMatrix();
+
+        frameFields["frame.platform.minRawValue"] = frame.platform().minRawValue();
+        frameFields["frame.platform.maxRawValue"] = frame.platform().maxRawValue();
+
+        frameFields["frame.illuminant1"] = 3000;        
+        frameFields["frame.colorMatrix1"] = rawToRGB3000;
+        frameFields["frame.illuminant2"] = 6500;        
+        frameFields["frame.colorMatrix2"] = rawToRGB6500;
+                
+        // Write frame fields
+        for (TagMap::const_iterator it = frameFields.begin(); it != frameFields.end(); it++) {
+            privateData << TagValue(it->first).toBlob() << it->second.toBlob();
+        }
+
+        // Write tags
+        for (TagMap::const_iterator it = frame.tags().begin(); it != frame.tags().end(); it++) {
+            privateData << TagValue(it->first).toBlob() << it->second.toBlob();
+        }
+    }
 
     void saveDNG(Frame frame, const std::string &filename) {
         dprintf(DBG_MINOR, "saveDNG: Starting to write %s\n", filename.c_str());
@@ -81,7 +179,7 @@ namespace FCam {
                   "saveDNG: Cannot save a non-RAW frame as a DNG %s", filename.c_str());
             return;
         }
-        if (frame.bayerPattern() == NotBayer) {
+        if (frame.platform().bayerPattern() == NotBayer) {
             error(Event::FileSaveError, frame,
                   "saveDNG: Cannot save non-Bayer pattern RAW data as %s", filename.c_str());
             return;
@@ -90,14 +188,18 @@ namespace FCam {
         // Figure out the color matrices for this sensor
         std::vector<float> rawToRGB3000(12);
         std::vector<float> rawToRGB6500(12);
-        frame.rawToRGBColorMatrix(3000, &rawToRGB3000[0]);
-        frame.rawToRGBColorMatrix(6500, &rawToRGB6500[0]);
+        frame.platform().rawToRGBColorMatrix(3000, &rawToRGB3000.front());
+        frame.platform().rawToRGBColorMatrix(6500, &rawToRGB6500.front());
 
         // First map from raw to XYZ
-        double rawToXYZ3000[] = {0, 0, 0, 0, 0, 0, 0, 0, 0};
-        double rawToXYZ6500[] = {0, 0, 0, 0, 0, 0, 0, 0, 0};
+        std::vector<double> rawToXYZ3000(9);
+        std::vector<double> rawToXYZ6500(9);
+        std::vector<double> rawToRGB3000_linear(9);
+        std::vector<double> rawToRGB6500_linear(9);
         for (int i = 0; i < 3; i++) {
             for (int j = 0; j < 3; j++) {
+                rawToRGB3000_linear[i*3+j] = rawToRGB3000[i*4+j];
+                rawToRGB6500_linear[i*3+j] = rawToRGB6500[i*4+j];
                 for (int k = 0; k < 3; k++) {
                     rawToXYZ3000[i*3+j] += FCam::RGBtoXYZ[i*3+k]*rawToRGB3000[k*4+j];
                     rawToXYZ6500[i*3+j] += FCam::RGBtoXYZ[i*3+k]*rawToRGB6500[k*4+j];
@@ -108,28 +210,56 @@ namespace FCam {
         std::vector<double> xyzToRaw3000(9);
         std::vector<double> xyzToRaw6500(9);
         // Then invert
-        invert3x3(rawToXYZ3000, &xyzToRaw3000.front());
-        invert3x3(rawToXYZ6500, &xyzToRaw6500.front());
+        invert3x3(&rawToXYZ3000.front(), &xyzToRaw3000.front());
+        invert3x3(&rawToXYZ6500.front(), &xyzToRaw6500.front());
 
+        // Need to back-calculate black levels as raw sensor values as well
+        // Black levels may vary as a function of illuminant (glare?), so calculate black level as a function of frame white balance
+        std::vector<double> rgbToRAW3000(9);
+        std::vector<double> rgbToRAW6500(9);
+        invert3x3(&rawToRGB3000_linear.front(), &rgbToRAW3000.front() );
+        invert3x3(&rawToRGB6500_linear.front(), &rgbToRAW6500.front() );
+        std::vector<double> blackLevel3000(3), blackLevel6500(3);
+        for (int i=0; i < 3; i++) {
+            for (int j=0; j < 3; j++) {
+                blackLevel3000[i] += rgbToRAW3000[i*3+j] * rawToRGB3000[3+j*4];
+                blackLevel6500[i] += rgbToRAW6500[i*3+j] * rawToRGB6500[3+j*4];
+            }
+        }
+        // linear interpolation with inverse correlated color temperature
+        float wbInv = 1.0/frame.whiteBalance();
+        float kInv3000 = 1.0/3000.f;
+        float kInv6500 = 1.0/6500.f;
+
+        float alpha = (wbInv - kInv6500)/(kInv3000-kInv6500);
+        std::vector<int> blackLevel(3);
+        for (int i=0; i < 3; i++) {
+            blackLevel[i] = static_cast<int>(-std::floor(alpha*blackLevel6500[i]+(1-alpha)*blackLevel3000[i]+0.5));
+        }
+        dprintf(5, "saveDNG: Black level 6500k: %f %f %f, 3000k: %f %f %f, WB: %d k, alpha: %f, final black level: %d %d %d\n",
+                blackLevel6500[0],blackLevel6500[1],blackLevel6500[2],
+                blackLevel3000[0],blackLevel3000[1],blackLevel3000[2],
+                frame.whiteBalance(),
+                alpha,
+                blackLevel[0], blackLevel[1], blackLevel[2]);
+  
         // Start constructing DNG fields
 
-        //bool success;
+        TiffFile dng;
 
-        TIFFFile dng;
-
-        TIFFFile::Ifd *ifd0 = dng.addIfd();
+        TiffIfd *ifd0 = dng.addIfd();
 
         // Add IFD0 entries
 
         std::string dngVersion(understoodDNGVersion,4);
         ifd0->add(DNG_TAG_DNGVersion, dngVersion);
-
+        
         std::string dngBackVersion(oldestSupportedDNGVersion,4);
         ifd0->add(DNG_TAG_DNGBackwardVersion, dngBackVersion);
 
-        ifd0->add(TIFF_TAG_Make, frame.manufacturer());
-        ifd0->add(TIFF_TAG_Model, frame.model());
-        ifd0->add(DNG_TAG_UniqueCameraModel, frame.model());
+        ifd0->add(TIFF_TAG_Make, frame.platform().manufacturer());
+        ifd0->add(TIFF_TAG_Model, frame.platform().model());
+        ifd0->add(DNG_TAG_UniqueCameraModel, frame.platform().model());
 
         if (frame.tags().find("flash.brightness") != frame.tags().end()) {
             // \todo Find actual spec on this, implement better
@@ -142,15 +272,15 @@ namespace FCam {
             ifd0->add(TIFFEP_TAG_Flash, 1);
         }
 
-        // \todo Find specs on this one
-        ifd0->add(TIFFEP_TAG_TIFFEPStandardID, 1);
+        std::string tiffEPStandardID(tiffEPVersion, 4);
+        ifd0->add(TIFFEP_TAG_TIFFEPStandardID, tiffEPStandardID);
 
         time_t tim = frame.exposureStartTime().s();
         struct tm *local = localtime(&tim);
         char buf[20];
         snprintf(buf, 20, "%04d:%02d:%02d %02d:%02d:%02d",
                  local->tm_year + 1900,
-                 local->tm_mon,
+                 local->tm_mon+1,
                  local->tm_mday,
                  local->tm_hour,
                  local->tm_min,
@@ -163,6 +293,8 @@ namespace FCam {
         ifd0->add(DNG_TAG_CalibrationIlluminant2, DNG_TAG_CalibrationIlluminant_D65);
         ifd0->add(DNG_TAG_ColorMatrix2, xyzToRaw6500);
 
+        ifd0->add(TIFF_TAG_Orientation, TIFF_Orientation_TopLeft);
+
         std::vector<double> whiteXY(2);
         float x,y;
         kelvinToXY(frame.whiteBalance(), &x, &y);
@@ -179,12 +311,26 @@ namespace FCam {
         }
         
         // Add some EXIF tags
-        ifd0->add(EXIF_TAG_ExposureTime, double(frame.exposure())/1e6);
+        TiffIfd *exifIfd = ifd0->addExifIfd();
+
+        exifIfd->add(EXIF_TAG_ExposureTime, double(frame.exposure())/1e6);
         if (frame.tags().find("lens.aperture") != frame.tags().end()) {
             double fNumber = frame["lens.aperture"].asFloat();
             ifd0->add(EXIF_TAG_FNumber, fNumber);
         }
 
+        int usecs = frame.exposureStartTime().us();
+        snprintf(buf, 20, "%06d", usecs);
+        exifIfd->add(EXIF_TAG_SubsecTime, std::string(buf));
+
+        exifIfd->add(EXIF_TAG_SensitivityType, EXIF_TAG_SensitivityType_ISO);
+        exifIfd->add(EXIF_TAG_ISOSpeedRatings, (int)(frame.gain()*100) );
+
+        if (frame.tags().find("lens.zoom") != frame.tags().end()) {
+            double focalLength = frame["lens.zoom"].asFloat();
+            exifIfd->add(EXIF_TAG_FocalLength, focalLength);
+        }
+
         // Create our very own DNG private data!
         {
             std::stringstream privateData;
@@ -192,48 +338,29 @@ namespace FCam {
             privateData << privateDataPreamble << std::ends;
             // never remove this
             privateData << TagValue(privateDataVersion).toBlob();
-            // if you change the order/count of the below, make sure to update the loading side as well, and up the version number.
-            privateData << TagValue(frame.exposureStartTime()).toBlob();
-            privateData << TagValue(frame.exposureEndTime()).toBlob();
-            privateData << TagValue(frame.processingDoneTime()).toBlob();
-            privateData << TagValue(frame.exposure()).toBlob();
-            privateData << TagValue(frame.frameTime()).toBlob();
-            privateData << TagValue(frame.gain()).toBlob();
-            privateData << TagValue(frame.whiteBalance()).toBlob();
-
-            privateData << TagValue(frame.shot().exposure).toBlob();
-            privateData << TagValue(frame.shot().frameTime).toBlob();
-            privateData << TagValue(frame.shot().gain).toBlob();
-            privateData << TagValue(frame.shot().whiteBalance).toBlob();
-
-            privateData << TagValue(frame.minRawValue()).toBlob();
-            privateData << TagValue(frame.maxRawValue()).toBlob();
-
-            privateData << TagValue(3000).toBlob();
-            privateData << TagValue(rawToRGB3000).toBlob();
-            privateData << TagValue(6500).toBlob();
-            privateData << TagValue(rawToRGB6500).toBlob();
-
-            // (not writing histogram/sharpness map)
-
-
-            // Write tags
-            for (TagMap::const_iterator it = frame.tags().begin(); it != frame.tags().end(); it++) {
-                privateData << TagValue(it->first) << TagValue(it->second);
+            switch (privateDataVersion) {
+            case 1:
+                saveDNGPrivateData_v1(frame, privateData, rawToRGB3000, rawToRGB6500);
+                break;
+            case 2:
+                saveDNGPrivateData_v2(frame, privateData, rawToRGB3000, rawToRGB6500);
+                break;
             }
             ifd0->add(DNG_TAG_DNGPrivateData, privateData.str());
         }
         // Add thumbnail into thumbnail IFD
 
-        TIFFFile::Ifd *thumbIfd = ifd0;
+        dprintf(4, "saveDNG: Adding thumbnail\n");
+        TiffIfd *thumbIfd = ifd0;
         Image thumbnail = makeThumbnail(frame);
 
         thumbIfd->add(TIFF_TAG_NewSubFileType, (int)TIFF_NewSubfileType_MainPreview);
         thumbIfd->setImage(thumbnail);
 
         // Add RAW Ifd and entries
+        dprintf(4, "saveDNG: Adding RAW metadata\n");
 
-        TIFFFile::Ifd *rawIfd = ifd0->addSubIfd();
+        TiffIfd *rawIfd = ifd0->addSubIfd();
 
         rawIfd->add(TIFF_TAG_NewSubFileType, (int)TIFF_NewSubfileType_FullRAW);
 
@@ -241,59 +368,203 @@ namespace FCam {
         rawIfd->add(TIFFEP_TAG_CFARepeatPatternDim, CFARepeatPatternDim);
 
         std::string CFAPattern;
-        std::vector<int> blackLevel;
-        switch(frame.bayerPattern()) {
+        std::vector<int> blackLevelPattern;
+        switch(frame.platform().bayerPattern()) {
         case RGGB:
             CFAPattern = std::string(TIFFEP_CFAPattern_RGGB,4);
-            blackLevel.push_back(-rawToRGB3000[3]);
-            blackLevel.push_back(-rawToRGB3000[7]);
-            blackLevel.push_back(-rawToRGB3000[7]);
-            blackLevel.push_back(-rawToRGB3000[11]);
+            blackLevelPattern.push_back(blackLevel[0]);
+            blackLevelPattern.push_back(blackLevel[1]);
+            blackLevelPattern.push_back(blackLevel[1]);
+            blackLevelPattern.push_back(blackLevel[2]);
             break;
         case BGGR:
             CFAPattern = std::string(TIFFEP_CFAPattern_BGGR,4);
-            blackLevel.push_back(-rawToRGB3000[11]);
-            blackLevel.push_back(-rawToRGB3000[7]);
-            blackLevel.push_back(-rawToRGB3000[7]);
-            blackLevel.push_back(-rawToRGB3000[3]);
+            blackLevelPattern.push_back(blackLevel[2]);
+            blackLevelPattern.push_back(blackLevel[1]);
+            blackLevelPattern.push_back(blackLevel[1]);
+            blackLevelPattern.push_back(blackLevel[0]);
             break;
         case GRBG:
             CFAPattern = std::string(TIFFEP_CFAPattern_GRBG,4);
-            blackLevel.push_back(-rawToRGB3000[7]);
-            blackLevel.push_back(-rawToRGB3000[3]);
-            blackLevel.push_back(-rawToRGB3000[11]);
-            blackLevel.push_back(-rawToRGB3000[7]);
+            blackLevelPattern.push_back(blackLevel[1]);
+            blackLevelPattern.push_back(blackLevel[0]);
+            blackLevelPattern.push_back(blackLevel[2]);
+            blackLevelPattern.push_back(blackLevel[1]);
             break;
         case GBRG:
             CFAPattern = std::string(TIFFEP_CFAPattern_GBRG,4);
-            blackLevel.push_back(-rawToRGB3000[7]);
-            blackLevel.push_back(-rawToRGB3000[11]);
-            blackLevel.push_back(-rawToRGB3000[3]);
-            blackLevel.push_back(-rawToRGB3000[7]);
+            blackLevelPattern.push_back(blackLevel[1]);
+            blackLevelPattern.push_back(blackLevel[2]);
+            blackLevelPattern.push_back(blackLevel[0]);
+            blackLevelPattern.push_back(blackLevel[1]);
         default:
             error(Event::FileSaveError, "saveDNG: %s: Can't handle non-bayer RAW images", filename.c_str());
             return;
             break;
         }
         rawIfd->add(TIFFEP_TAG_CFAPattern, CFAPattern);
-        rawIfd->add(DNG_TAG_BlackLevel, blackLevel);
+        rawIfd->add(DNG_TAG_BlackLevel, blackLevelPattern);
 
         std::vector<int> blackLevelRepeatDim(2,2);
         rawIfd->add(DNG_TAG_BlackLevelRepeatDim, blackLevelRepeatDim);
 
-        rawIfd->add(DNG_TAG_WhiteLevel, frame.maxRawValue());
+        rawIfd->add(DNG_TAG_WhiteLevel, frame.platform().maxRawValue());
 
+        // Leaving a 4-pixel border around RAW image to allow for loss of data during interpolation
+        std::vector<int> cropOrigin(2);
+        cropOrigin[0] = 4;
+        cropOrigin[1] = 4;
+        std::vector<int> cropSize(2);
+        cropSize[0] = frame.image().width()-8;
+        cropSize[1] = frame.image().height()-8;
+
+        rawIfd->add(DNG_TAG_DefaultCropOrigin, cropOrigin);
+        rawIfd->add(DNG_TAG_DefaultCropSize, cropSize);
+
+        dprintf(4, "saveDNG: Adding RAW image\n");
         rawIfd->setImage(frame.image());
 
+        dprintf(4, "saveDNG: Beginning write to disk\n");
         // Constructed all DNG fields, write it to disk
         dng.writeTo(filename);
 
         dprintf(DBG_MINOR, "saveDNG: Done writing %s\n", filename.c_str());
     }
 
+    void loadDNGPrivateData_v1(_DNGFrame *_f, std::stringstream &privateData) {
+        // Deserialize our fields now
+        // Order below must match DNG save order
+        TagValue val;
+        privateData >> val;
+        _f->exposureStartTime = val;
+        privateData >> val;
+        _f->exposureEndTime = val;
+        privateData >> val;
+        _f->processingDoneTime = val;
+        privateData >> val;
+        _f->exposure = val;
+        privateData >> val;
+        _f->frameTime = val;
+        privateData >> val;
+        _f->gain = val;
+        privateData >> val;
+        _f->whiteBalance = val;
+        
+        privateData >> val;
+        _f->_shot.exposure = val;
+        privateData >> val;
+        _f->_shot.frameTime = val;
+        privateData >> val;
+        _f->_shot.gain = val;
+        privateData >> val;
+        _f->_shot.whiteBalance = val;
+        
+        privateData >> val;
+        _f->dng.minRawValue = val.asInt();
+        privateData >> val;
+        _f->dng.maxRawValue = val.asInt();
+        
+        _f->dng.numIlluminants = 2;
+        
+        privateData >> val;
+        _f->dng.illuminant1 = val;
+        privateData >> val;
+        for (int i=0; i < 12 ; i++)
+            _f->dng.colorMatrix1[i] = val.asFloatVector()[i];
+        
+        privateData >> val;
+        _f->dng.illuminant2 = val;
+        privateData >> val;
+        for (int i=0; i < 12 ; i++)
+            _f->dng.colorMatrix2[i] = val.asFloatVector()[i];
+        
+        // And the tags
+        TagValue key;
+        while ((privateData >> key).good()) {
+            privateData >> val;
+            _f->tags[key] = val;
+        }
+    }
+
+    void loadDNGPrivateData_v2(_DNGFrame *_f, std::stringstream &privateData) {
+
+        // Read everything into the frame tags
+        TagValue key, val;
+        while ((privateData >> key).good()) {
+            privateData >> val;
+            _f->tags[key] = val;
+        }
+
+        // And then look for special frame fields and parse them out
+
+        TagMap::iterator it;
+
+#define grabField(fieldName, fieldDest, asType)      \
+        do {                                         \
+            it = _f->tags.find(fieldName);           \
+            if ( it != _f->tags.end()) {             \
+                fieldDest = it->second.asType;       \
+                _f->tags.erase(it);                  \
+            }                                        \
+        } while(0)
+
+        grabField("frame.exposureStartTime", _f->exposureStartTime, asTime());
+        grabField("frame.exposureEndTime", _f->exposureEndTime, asTime());
+
+        grabField("frame.processingDoneTime", _f->processingDoneTime, asTime());
+        grabField("frame.exposure", _f->exposure, asInt());
+        grabField("frame.frameTime", _f->frameTime, asInt());
+        grabField("frame.gain", _f->gain, asFloat());
+
+        grabField("frame.whiteBalance", _f->whiteBalance, asInt());
+
+        grabField("frame.shot.exposure", _f->_shot.exposure, asInt());
+        grabField("frame.shot.frameTime", _f->_shot.frameTime, asInt());
+        grabField("frame.shot.gain", _f->_shot.gain, asFloat());
+
+        grabField("frame.shot.whiteBalance", _f->_shot.whiteBalance, asInt());
+        
+        std::vector<float> cMatrix;
+        grabField("frame.shot.colorMatrix", cMatrix, asFloatVector());
+        _f->_shot.setColorMatrix(cMatrix);
+
+        grabField("frame.platform.minRawValue", _f->dng.minRawValue, asInt());
+        grabField("frame.platform.maxRawValue", _f->dng.maxRawValue, asInt());
+
+        _f->dng.numIlluminants = 2;
+        grabField("frame.illuminant1", _f->dng.illuminant1, asInt());
+        grabField("frame.illuminant2", _f->dng.illuminant2, asInt());
+
+        it = _f->tags.find("frame.colorMatrix1");
+        if (it != _f->tags.end()) {
+            for (int i=0; i < 12; i++) {
+                _f->dng.colorMatrix1[i] = it->second.asFloatVector()[i];
+            }
+            _f->tags.erase(it);
+        }
+
+        it = _f->tags.find("frame.colorMatrix2");
+        if (it != _f->tags.end()) {
+            for (int i=0; i < 12; i++) {
+                _f->dng.colorMatrix2[i] = it->second.asFloatVector()[i];
+            }
+            _f->tags.erase(it);
+        }
+
+        // All leftover tags stay in the frame tags
+    }
+
 
     DNGFrame loadDNG(const std::string &filename) {
-        TIFFFile dng(filename);
+        // Construct DNG Frame
+        _DNGFrame *_f = new _DNGFrame;
+        DNGFrame f(_f);
+        // We'll mostly manipulate the internal dngFile for now
+        TiffFile &dng = *(_f->dngFile);
+
+        // Start reading DNG file data
+        dng.readFrom(filename);
+
         if (!dng.valid) {
             // An error during initial TIFF parsing is recorded in
             // dng.lastEvent, forward it to the event system
@@ -304,7 +575,7 @@ namespace FCam {
             return nullF;
         }
 
-        // Convenience macro, used a lot below.
+        // Convenience error macro
 #define fatalError(...) do {                                            \
             DNGFrame nullF;                                             \
             error(Event::FileLoadError, nullF, __VA_ARGS__);            \
@@ -313,7 +584,7 @@ namespace FCam {
         //
         // Verify DNG version
 
-        const TIFFFile::IfdEntry *versionEntry, *backVersionEntry;
+        const TiffIfdEntry *versionEntry, *backVersionEntry;
 
         versionEntry = dng.ifds(0)->find(DNG_TAG_DNGVersion);
         if (!versionEntry) fatalError("loadDNG: File %s is not a valid DNG file (no DNG version found)", filename.c_str());
@@ -343,14 +614,14 @@ namespace FCam {
         //
         // Let's find the RAW IFD
 
-        const TIFFFile::IfdEntry *entry;
+        const TiffIfdEntry *entry;
         entry = dng.ifds(0)->find(TIFF_TAG_NewSubFileType);
         int ifdType = TIFF_NewSubfileType_DEFAULT;
         if (entry) {
             ifdType = entry->value();
         }
 
-        TIFFFile::Ifd *rawIfd=NULL;
+        TiffIfd *rawIfd = NULL;
         if (ifdType == (int)TIFF_NewSubfileType_FullRAW) {
             // Main IFD has RAW data
             dprintf(4, "loadDNG: RAW data found in IFD0\n");
@@ -373,15 +644,11 @@ namespace FCam {
         }
 
         //
-        // Ok, now to parse the RAW data
-
-        // Get the bare image data pointed to by the RAW ifd
-        Image img = rawIfd->getImage();
-        if (!img.valid()) {
-            fatalError("loadDNG: %s: Cannot load RAW image data", filename.c_str());
-        }
-
-        // Need a bunch of other metadata
+        // Read in RAW image data 
+        _f->image = rawIfd->getImage();
+        
+        //
+        // Ok, now to parse the RAW metadata
 
         // Read in Bayer mosaic pattern
         entry = rawIfd->find(TIFFEP_TAG_CFARepeatPatternDim);
@@ -495,6 +762,7 @@ namespace FCam {
 
         //
         // Let's find the Thumbnail IFD, if any
+        TiffIfd *thumbIfd = NULL;
 
         entry = dng.ifds(0)->find(TIFF_TAG_NewSubFileType);
         ifdType = TIFF_NewSubfileType_DEFAULT;
@@ -502,7 +770,6 @@ namespace FCam {
             ifdType = entry->value();
         }
 
-        TIFFFile::Ifd *thumbIfd=NULL;
         if (ifdType == (int)TIFF_NewSubfileType_MainPreview) {
             // Main IFD has thumbnail data
             dprintf(4, "loadDNG: Thumbnail data found in IFD0\n");
@@ -522,15 +789,21 @@ namespace FCam {
                 }
             }            
         }
-        Image thumbnail;
+
         if (thumbIfd) {
-            dprintf(4, "loadDNG: %s: Loading thumbnail\n", filename.c_str());
-            thumbnail = thumbIfd->getImage();
+            _f->thumbnail = thumbIfd->getImage();
         }
 
         //
         // Let's grab other useful metadata from the main IFD
 
+        // Verify orientation
+        entry = dng.ifds(0)->find(TIFF_TAG_Orientation);
+        if (!entry) fatalError("loadDNG: %s: No orientation entry found", filename.c_str());
+        int orientation = entry->value();
+
+        if (orientation != TIFF_Orientation_TopLeft) fatalError("loadDNG: %s: Unsupported orientation value %d", filename.c_str(), orientation);
+
         // Read in color calibration information
         entry = dng.ifds(0)->find(DNG_TAG_CalibrationIlluminant1);
         unsigned int calibIlluminant1 = 0;
@@ -541,6 +814,7 @@ namespace FCam {
         dprintf(4,"loadDNG: %s: Calibration illuminant #1 is type %d (%d K)\n", filename.c_str(), calibIlluminant1, calibTemp1);
 
         std::vector<float> camToRGBMatrix1(9);
+        std::vector<float> rgbBlackLevel1(3);
         {
             entry = dng.ifds(0)->find(DNG_TAG_ColorMatrix1);
             if (!entry) fatalError("loadDNG: %s: No color calibration matrix found in IFD0", filename.c_str());
@@ -557,12 +831,18 @@ namespace FCam {
                 }
             }
             invert3x3(&rgbToCamMatrix1[0], &camToRGBMatrix1[0]);
+            for (int i=0;i<3;i++) {
+                for (int j=0; j<3; j++) {
+                    rgbBlackLevel1[i] += camToRGBMatrix1[i*3+j]*(-blackLevel[j]);
+                }
+            }
         }
 
         entry = dng.ifds(0)->find(DNG_TAG_CalibrationIlluminant2);
         int numCalibIlluminants = 1;
         int calibTemp2 = -1;
         std::vector<float> camToRGBMatrix2(9);
+        std::vector<float> rgbBlackLevel2(3);
         if (entry) {
             numCalibIlluminants = 2;
             unsigned int calibIlluminant2 = (int)entry->value();
@@ -577,7 +857,6 @@ namespace FCam {
             std::vector<double> &xyzToCamMatrix2 = entry->value();
 
             std::vector<float> rgbToCamMatrix2(9);
-
             for (int i=0;i<3;i++) {
                 for (int j=0; j<3; j++) {
                     for (int k=0; k<3;k++) {
@@ -586,6 +865,12 @@ namespace FCam {
                 }
             }
             invert3x3(&rgbToCamMatrix2[0], &camToRGBMatrix2[0]);
+            for (int i=0;i<3;i++) {
+                for (int j=0; j<3; j++) {
+                    rgbBlackLevel2[i] += camToRGBMatrix2[i*3+j]*(-blackLevel[j]);
+                }
+            }
+
         }
 
         // Read in camera model
@@ -602,10 +887,6 @@ namespace FCam {
 
 #undef fatalError
 
-        // Create the frame in case we're sticking private data into it
-        _DNGFrame *_f = new _DNGFrame;
-        DNGFrame DNGFrame(_f);
-
         // Write values that might be overriden by private data
         _f->dng.minRawValue = std::min(blackLevel[0], std::min(blackLevel[1], blackLevel[2]));
         _f->dng.maxRawValue = whiteLevel;
@@ -619,8 +900,8 @@ namespace FCam {
             }
         }
         for (int i=0; i < 3; i++) {
-            _f->dng.colorMatrix1[i*4+3] = -blackLevel[i];
-            _f->dng.colorMatrix2[i*4+3] = -blackLevel[i];
+            _f->dng.colorMatrix1[i*4+3] = rgbBlackLevel1[i];
+            _f->dng.colorMatrix2[i*4+3] = rgbBlackLevel2[i];
         }
 
         // Grab our private DNG data to supplement/replace the above
@@ -635,64 +916,23 @@ namespace FCam {
                 std::stringstream privateData(privateString.substr(preambleEnd+1));
                 TagValue version;
                 privateData >> version;
-                if (version.asInt() == privateDataVersion) {
-                    dprintf(4,"loadDNG: %s: Reading private data: %s version %d.\n", filename.c_str(), preamble.c_str(), version.asInt());
-                    // Deserialize our fields now
-                    // Order below must match DNG save order
-                    TagValue val;
-                    privateData >> val;
-                    _f->exposureStartTime = val;
-                    privateData >> val;
-                    _f->exposureEndTime = val;
-                    privateData >> val;
-                    _f->processingDoneTime = val;
-                    privateData >> val;
-                    _f->exposure = val;
-                    privateData >> val;
-                    _f->frameTime = val;
-                    privateData >> val;
-                    _f->gain = val;
-                    privateData >> val;
-                    _f->whiteBalance = val;
-
-                    privateData >> val;
-                    _f->_shot.exposure = val;
-                    privateData >> val;
-                    _f->_shot.frameTime = val;
-                    privateData >> val;
-                    _f->_shot.gain = val;
-                    privateData >> val;
-                    _f->_shot.whiteBalance = val;
-
-                    privateData >> val;
-                    _f->dng.minRawValue = val.asInt();
-                    privateData >> val;
-                    _f->dng.maxRawValue = val.asInt();
-
-                    _f->dng.numIlluminants = 2;
-
-                    privateData >> val;
-                    _f->dng.illuminant1 = val;
-                    privateData >> val;
-                    for (int i=0; i < 12 ; i++)
-                        _f->dng.colorMatrix1[i] = val.asFloatVector()[i];
-
-                    privateData >> val;
-                    _f->dng.illuminant2 = val;
-                    privateData >> val;
-                    for (int i=0; i < 12 ; i++)
-                        _f->dng.colorMatrix2[i] = val.asFloatVector()[i];
-
-                    // And the tags
-                    TagValue key;
-                    while ((privateData >> key).good()) {
-                        privateData >> val;
-                        _f->tags[key] = val;
-                    }
+                
+                if (version.asInt() == 1) {
+                    dprintf(4,"loadDNG: %s: Reading private data, version 1.\n", filename.c_str());
+                    loadDNGPrivateData_v1(_f, privateData);
                 } else {
-                    warning(Event::FileLoadError,
-                            "loadDNG: %s: Private data version too new: %d (can handle %d), ignoring it.",
-                            filename.c_str(), version.asInt(), privateDataVersion);
+                    TagValue backwardVersion;
+                    privateData >> backwardVersion;
+                    switch (backwardVersion.asInt()) {
+                    case 2:
+                        dprintf(4,"loadDNG: %s: Reading private data, version 2.\n", filename.c_str());
+                        loadDNGPrivateData_v2(_f, privateData);
+                        break;                        
+                    default:
+                        warning(Event::FileLoadError,
+                                "loadDNG: %s: Private data version too new: %d,%d (can handle X, %d), ignoring it.",
+                                filename.c_str(), version.asInt(), backwardVersion.asInt(), privateDataVersion);
+                    }
                 }
             } else {
                 warning(Event::FileLoadError,
@@ -702,14 +942,12 @@ namespace FCam {
         }
         // Write remaining frame information
 
-        _f->image = img;
-        _f->thumbnail = thumbnail;
         _f->dng.bayerPattern = bayer;
         _f->dng.manufacturer = cameraManufacturer;
         _f->dng.model = cameraModel;
 
         dprintf(4,"loadDNG: %s: Loaded successfully\n", filename.c_str());
-        return DNGFrame;
+        return f;
     }
 
 }
