@@ -17,8 +17,8 @@ namespace FCam {
 
     // Make a linear luminance -> pixel value lookup table
     void makeLUT(const Frame &f, float contrast, int blackLevel, float gamma, unsigned char *lut) {
-        unsigned short minRaw = f.minRawValue()+blackLevel;
-        unsigned short maxRaw = f.maxRawValue();
+        unsigned short minRaw = f.platform().minRawValue()+blackLevel;
+        unsigned short maxRaw = f.platform().maxRawValue();
         
         for (int i = 0; i <= minRaw; i++) {
             lut[i] = 0;
@@ -76,7 +76,7 @@ namespace FCam {
         Image input = src.image();
 
         // First check we're the right bayer pattern. If not crop and continue.
-        switch((int)src.bayerPattern()) {
+        switch((int)src.platform().bayerPattern()) {
         case GRBG:
             break;
         case RGGB:
@@ -126,7 +126,15 @@ namespace FCam {
 
         // Grab the color matrix
         float colorMatrix[12];
-        src.rawToRGBColorMatrix(colorMatrix);
+        // Check if there's a custom color matrix
+        if (src.shot().colorMatrix().size() == 12) {
+            for (int i = 0; i < 12; i++) {
+                colorMatrix[i] = src.shot().colorMatrix()[i];
+            }
+        } else {
+            // Otherwise use the platform version
+            src.platform().rawToRGBColorMatrix(src.shot().whiteBalance, colorMatrix);
+        }
 
         for (int by = 0; by < rawHeight-8-BLOCK_HEIGHT+1; by += BLOCK_HEIGHT) {
             for (int bx = 0; bx < rawWidth-8-BLOCK_WIDTH+1; bx += BLOCK_WIDTH) {
@@ -455,7 +463,7 @@ namespace FCam {
             src.image().height() == 1968 &&
             thumbSize.width == 640 &&
             thumbSize.height == 480 &&
-            src.bayerPattern() == GRBG) {
+            src.platform().bayerPattern() == GRBG) {
             return makeThumbnailRAW_ARM(src, contrast, blackLevel, gamma);
         }
         #endif
@@ -467,7 +475,7 @@ namespace FCam {
         Image thumb;
         //!TODO: Add thumbnailing support for other Bayer patterns 
         bool redRowEven, blueRowGreenPixelEven;
-        switch (src.bayerPattern()) {
+        switch (src.platform().bayerPattern()) {
         case RGGB:
             redRowEven = true;
             blueRowGreenPixelEven = true;
@@ -509,7 +517,15 @@ namespace FCam {
         if (cropY % 2 == 1) cropY--; // Ensure we're at start of 2x2 block
 
         float colorMatrix[12];
-        src.rawToRGBColorMatrix(colorMatrix);
+        // Check if there's a custom color matrix
+        if (src.shot().colorMatrix().size() == 12) {
+            for (int i = 0; i < 12; i++) {
+                colorMatrix[i] = src.shot().colorMatrix()[i];
+            }
+        } else {
+            // Otherwise use the platform version
+            src.platform().rawToRGBColorMatrix(src.shot().whiteBalance, colorMatrix);
+        }
 
         /* A fast downsampling/demosaicing - average down color
            channels over the whole source pixel block under a single
