@@ -15,7 +15,7 @@ namespace FCam {
         Image input = src.image();
 
         // Check we're the right bayer pattern. If not crop and continue.
-        switch((int)src.bayerPattern()) {
+        switch((int)src.platform().bayerPattern()) {
         case GRBG:
             break;
         case RGGB:
@@ -66,7 +66,15 @@ namespace FCam {
         // Prepare the color matrix in S8.8 fixed point
         float colorMatrix_f[12];
         
-        src.rawToRGBColorMatrix((float *)colorMatrix_f);
+        // Check if there's a custom color matrix
+        if (src.shot().colorMatrix().size() == 12) {
+            for (int i = 0; i < 12; i++) {
+                colorMatrix_f[i] = src.shot().colorMatrix()[i];
+            }
+        } else {
+            // Otherwise use the platform version
+            src.platform().rawToRGBColorMatrix(src.shot().whiteBalance, colorMatrix_f);
+        }
 
         int16x4_t colorMatrix[3];
         for (int i = 0; i < 3; i++) {
@@ -711,7 +719,18 @@ namespace FCam {
 
         Time startTime = Time::now();
         float colorMatrix_f[12];
-        src.rawToRGBColorMatrix(colorMatrix_f);
+
+        // Check if there's a custom color matrix
+        if (src.shot().colorMatrix().size() == 12) {
+            for (int i = 0; i < 12; i++) {
+                colorMatrix_f[i] = src.shot().colorMatrix()[i];
+            }
+            printf("Making thumbnail with custom WB\n");
+        } else {
+            // Otherwise use the platform version
+            src.platform().rawToRGBColorMatrix(src.shot().whiteBalance, colorMatrix_f);
+            printf("Making thumbnail with platform WB\n");
+        }
 
         register int16x4_t colorMatrix0 asm ("d0"); // ASM assignments are essential - they're implicitly trusted by the inline code.
         register int16x4_t colorMatrix1 asm ("d1");
