@@ -1180,6 +1180,8 @@ namespace FCam {
 
         if (success) success = add(TIFF_TAG_StripOffsets, stripOffsets);
         if (success) success = add(TIFF_TAG_StripByteCounts, stripByteCounts);
+        // Not needed per spec, but dcraw seems to need this for thumbnails
+        if (success) success = add(TIFF_TAG_Compression, TIFF_Compression_Uncompressed);
 
         if (!success) {
             error(Event::FileSaveError,
